// C Source File
// Created 08/06/2006; 10:31:11

#include <math.h>				// for 'abs'
#include <sprites.h>		// for 'SPRT_OR'
#include "prototypes.h"

// Important !
// Be careful not to use these functions out of your screen: there is no clipping !
// Parameter 'width': width of your original picture in BYTES
// Parameter 'height': height of your original picture in PIXELS
// Parameter 'dest': its width is supposed to be 30 bytes long (no checking is done ! so be careful with virtual screens)
// For 'FromPic_Sprite8', 'FromPic_Sprite16', 'FromPic_Sprite32', 'FromPic_DrawLineFast' and 'FromPic_Dissolve',
// only one drawing mode: RPLC (replace)

// These 3 following routines have been adapted from the TIGCC's ones (available in the sources)

//===============================================================================================================//

void FromPic_Sprite8(short x, short y, short h, const unsigned char * sprite, void * src, void * dest, short width)
{
  long addr_d = (long)dest+30*y+((x>>3)&0xFFFE), addr_s = (long)src+width*y+((x>>3)&0xFFFE), d1;
  unsigned short cnt = 24-(x&15), data;
  
  for(; h --; addr_d += 30, addr_s += width)
  {
    data = *(sprite ++);
    *(long *)addr_d &= ~((long)data<<cnt);
    d1 = *(long *)addr_s&((long)data<<cnt);
    *(long *)addr_d |= d1;
  }
}

//===============================================================================================================//

void FromPic_Sprite16(short x, short y, short h, const unsigned short * sprite, void * src, void * dest, short width)
{
  long addr_d = (long)dest+30*y+((x>>3)&0xFFFE), addr_s = (long)src+width*y+((x>>3)&0xFFFE), d1;
  unsigned short cnt = 16-(x&15), data;
  
  for(; h --; addr_d += 30, addr_s += width)
  {
    data = *(sprite ++);
    *(long *)addr_d &= ~((long)data<<cnt);
    d1 = *(long *)addr_s&((long)data<<cnt);
    *(long *)addr_d |= d1;
  }
}

//===============================================================================================================//

void FromPic_Sprite32(short x, short y, short h, const unsigned long * sprite, void * src, void * dest, short width)
{
  long addr_d = (long)dest+30*y+((x>>3)&0xFFFE), addr_s = (long)src+width*y+((x>>3)&0xFFFE);
  unsigned long data, d1, d2;
  short cnt = x&15, ccnt = 32-cnt;
  
  for(; h --; addr_d += 30, addr_s += width)
  {
    data = *(sprite ++);
    *(long *)addr_d &= ~(data>>cnt);
    *(long *)(addr_d+4) &= ~(data<<ccnt);
    d1 = *(long *)addr_s&(data>>cnt);
    d2 = *(long *)addr_s&(data<<ccnt);
    *(long *)addr_d |= d1;
    *(long *)(addr_d+4) |= d2;
  }
}

//===============================================================================================================//
//===============================================================================================================//

void FromPic_DrawLineFast(short x1, short y1, short x2, short y2, void * src, void * dest, short width)
{
  short x = x1, y = y1;
  short dx = abs(x2-x1), dy = abs(y2-y1);
  short ystep = ((y1 < y2) ? 1: -1), pystep_1 = 30*ystep, pystep_2 = width*ystep;
  short mov = (dx ? 0: -1);
  unsigned char * ptr_dest = (unsigned char *)(dest+30*y+(x>>3));
  unsigned char * ptr_src = (unsigned char *)(src+width*y+(x>>3));
  
  if (x1 < x2)
    while (x != x2 || y != y2)
    {
      *ptr_dest = (*ptr_dest&(~(1<<(7-(x&7)))))|(*ptr_src&(1<<(7-(x&7))));
      if (mov < 0) y += ystep, ptr_dest += pystep_1, ptr_src += pystep_2, mov += dx;
      else
      {
        mov -= dy;
        if(!(++ x & 7)) ptr_dest ++, ptr_src ++;
      }
    }
  else
    while (x != x2 || y != y2)
    {
      *ptr_dest = (*ptr_dest&(~(1<<(7-(x&7)))))|(*ptr_src&(1<<(7-(x&7))));
      if (mov < 0) y += ystep, ptr_dest += pystep_1, ptr_src += pystep_2, mov += dx;
      else
      {
        mov -= dy;
        if (!(x -- & 7)) ptr_dest --, ptr_src --;
      }
    }
}

//===============================================================================================================//

void FromPic_DrawHLineFast_OR(short x1, short x2, short y, void * src, void * dest, short width)
{
  short x = x1;
  short dx = abs(x2-x1);
  short mov = (dx ? 0: -1);
  unsigned char * ptr_dest = (unsigned char *)(dest+30*y+(x>>3));
  unsigned char * ptr_src = (unsigned char *)(src+width*y+(x>>3));
  unsigned short mask = 1<<(~x & 7);
  
  if (x1 < x2)
    while (x != x2)
    {
      *ptr_dest |= (*ptr_src&mask);
      if (mov < 0) mov += dx;
      else
      {
        if(++ x & 7) mask >>= 1;
        else ptr_dest ++, ptr_src ++, mask = 0x80;
      }
    }
  else
    while (x != x2)
    {
      *ptr_dest |= (*ptr_src&mask);
      if (mov < 0) mov += dx;
      else
      {
        if (x -- & 7) mask <<= 1;
        else ptr_dest --, ptr_src --, mask = 1;
      }
    }
}

//===============================================================================================================//

void FromPic_DrawHLineFast_RPLC(short x1, short x2, short y, void * src, void * dest, short width)
{
  short x = x1;
  short dx = abs(x2-x1);
  short mov = (dx ? 0: -1);
  unsigned char * ptr_dest = (unsigned char *)(dest+30*y+(x>>3));
  unsigned char * ptr_src = (unsigned char *)(src+width*y+(x>>3));
  
  if (x1 < x2)
    while (x != x2)
    {
      *ptr_dest = (*ptr_dest&(~(1<<(7-(x&7)))))|(*ptr_src&(1<<(7-(x&7))));
      if (mov < 0) mov += dx;
      else
      {
        if(!(++ x & 7)) ptr_dest ++, ptr_src ++;
      }
    }
  else
    while (x != x2)
    {
      *ptr_dest = (*ptr_dest&(~(1<<(7-(x&7)))))|(*ptr_src&(1<<(7-(x&7))));
      if (mov < 0) mov += dx;
      else
      {
        if (!(x -- & 7)) ptr_dest --, ptr_src --;
      }
    }
}

//===============================================================================================================//
// Author: Thomas Nussbaumer

void FromPic_Dissolve(void * src, void * dest, short width, short height)
{
  unsigned short seq = 1;
  unsigned char pixmask;
  short byteoffset;

  do
  {
  	pixmask = seq&0x7;
    byteoffset = seq>>3;
    
    if(byteoffset < height*width)
    {
    	pixmask = 0x80>>pixmask;
    	
    	if(*(unsigned char *)(src+byteoffset) & pixmask)
    	{
    		*(unsigned char *)(dest+30*(byteoffset/width)+byteoffset%width) |= pixmask;
    	}
    	else
    	{
    		*(unsigned char *)(dest+30*(byteoffset/width)+byteoffset%width) &= ~pixmask;
    	}
    }
    
    if(seq&1)
    {
    	seq = (seq>>1)^0x6000;
    }
    else
    {
    	seq = seq>>1;
    }

  }while(seq != 1);

}


//===============================================================================================================//

void FromPic_GrayDissolve(void * src_0, void * src_1, void * dest_0, void * dest_1, short width, short height)
{
  unsigned short seq = 1;
  unsigned char pixmask;
  short byteoffset;

  do
  {
  	pixmask = seq&0x7;
    byteoffset = seq>>3;
    
    if(byteoffset < height*width)
    {
    	pixmask = 0x80>>pixmask;
    	
    	if(*(unsigned char *)(src_0+byteoffset) & pixmask)
    	{
    		*(unsigned char *)(dest_0+30*(byteoffset/width)+byteoffset%width) |= pixmask;
    	}
    	else
    	{
    		*(unsigned char *)(dest_0+30*(byteoffset/width)+byteoffset%width) &= ~pixmask;
    	}
    	
    	if(*(unsigned char *)(src_1+byteoffset) & pixmask)
    	{
    		*(unsigned char *)(dest_1+30*(byteoffset/width)+byteoffset%width) |= pixmask;
    	}
    	else
    	{
    		*(unsigned char *)(dest_1+30*(byteoffset/width)+byteoffset%width) &= ~pixmask;
    	}
    	
    }
    
    if(seq&1)
    {
    	seq = (seq>>1)^0x6000;
    }
    else
    {
    	seq = seq>>1;
    }

  }while(seq != 1);

}

//===============================================================================================================//
//===============================================================================================================//

void FromPic_Disk(short radius, short x_center, short y_center, void * src, void * dest, short mode, short width)
{
 register short x, y, d;

 x = 0;
 y = radius;
 d = (5-radius)/4;
 
 FromPic_DrawLinesForDisk(x, y, y_center, x_center, src, dest, mode, width);
 
 while(y > x)
 {
 	if(d < 0)
 	{
 		d += x+x+3;
 	}
 	else
 	{
 		d += x+x-y-y+5;
 		y --;
 	}
 	
 	x ++;
 	FromPic_DrawLinesForDisk(x, y, y_center, x_center, src, dest, mode, width);
 }
 
}

//===============================================================================================================//

void FromPic_DrawLinesForDisk(short x, short y, short x_center, short y_center, void * src, void * dest, short mode, short width)
{
 if(mode == SPRT_OR)
 {
 	FromPic_DrawHLineFast_OR(y_center+y, y_center-y, x_center+x, src, dest, width);
 	FromPic_DrawHLineFast_OR(y_center+y, y_center-y, x_center-x, src, dest, width);
 	FromPic_DrawHLineFast_OR(y_center+x, y_center-x, x_center+y, src, dest, width);
 	FromPic_DrawHLineFast_OR(y_center+x, y_center-x, x_center-y, src, dest, width);
 }
 else if(mode == SPRT_RPLC)
 {
 	FromPic_DrawHLineFast_RPLC(y_center+y, y_center-y, x_center+x, src, dest, width);
 	FromPic_DrawHLineFast_RPLC(y_center+y, y_center-y, x_center-x, src, dest, width);
 	FromPic_DrawHLineFast_RPLC(y_center+x, y_center-x, x_center+y, src, dest, width);
 	FromPic_DrawHLineFast_RPLC(y_center+x, y_center-x, x_center-y, src, dest, width);
 }
}

//===============================================================================================================//
